import os
import aim
import argparse
import numpy as np
import jittor as jt
import optim as optim_temp 
from MMonitor.mmonitor.monitor import Monitor  
from MMonitor.visualize import Visualization
from jittor import nn
from jittor import models
import jittor.transform as transform
import copy
from data import *
import datetime
from PIL import Image
from jittor.dataset import Dataset
# from c2net.context import prepare,upload_output
# c2net_context = prepare()
parser = argparse.ArgumentParser(description='PyTorch WideResNet Training')
parser.add_argument('--model', default='resnet32', type=str, help='model')
parser.add_argument('--epochs', default=10, type=int, help='number of total epochs to run')
parser.add_argument('--batch_size', '--batch-size', default=100, type=int, help='mini-batch size (default: 100)')
parser.add_argument('--lr', '--learning-rate', default=1e-1, type=float, help='initial learning rate')
parser.add_argument('--momentum', default=0.9, type=float, help='momentum')
parser.add_argument('--weight-decay', '--wd', default=5e-4, type=float, help='weight decay (default: 5e-4)')
parser.add_argument('--prefetch', type=int, default=1, help='Pre-fetching threads.')
parser.add_argument('--seed', type=int, default=1)

args = parser.parse_args()
args.cifar_10_python_path = '/data/wlc/dataset/cifar' 
args.json_file_path = '/home/wlc/wlc/jittor-mmonitor/model'

args.sel_data = list(range(50000))

jt.flags.use_cuda = 1


def build_model():
    model = models.Resnet18(num_classes=args.num_classes)
    return model


def prepare_config():
    config_mmonitor_resnet18 = {
        nn.BatchNorm: ['ForwardInputSndNorm', 'ForwardInputMean', 'ForwardInputStd', 'WeightNorm','WeightMean']

    }
    return config_mmonitor_resnet18


def accuracy(output, target):
    batch_size = target.shape[0]
    pred = np.argmax(output, -1)
    res = ((pred == target).astype(float).sum()) / batch_size

    return res


def adjust_learning_rate(optimizer, epochs):
    lr = args.lr * ((0.1 ** int(epochs >= 80)) * (0.1 ** int(epochs >= 100)))  # For WRN-28-10
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def ce_loss(output, target, reduce=True):
    if len(output.shape) == 4:
        c_dim = output.shape[1]
        output = output.transpose((0, 2, 3, 1))
        output = output.reshape((-1, c_dim))
    target = target.reshape((-1,))
    target = target.broadcast(output, [1])
    target = target.index(1) == target

    output = output - output.max([1], keepdims=True)
    loss = output.exp().sum(1).log()
    loss = loss - (output * target).sum(1)
    if reduce:
        return loss.mean()
    else:
        return loss


def test(model, test_loader):
    model.eval()
    correct = 0
    test_loss = 0

    for _, (inputs, targets, _) in enumerate(test_loader):
        inputs, targets = jt.array(inputs), jt.array(targets)
        outputs = model(inputs)
        test_loss += nn.cross_entropy_loss(outputs, targets).detach().item()
        predicted = np.argmax(outputs.detach(), -1)
        correct += ((predicted == targets.data).astype(float).sum())

    test_loss /= len(test_loader)
    accuracy = 100. * correct / len(test_loader)

    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.4f}%)\n'.format(
        test_loss, correct, len(test_loader),
        accuracy))

    return accuracy


def train(train_loader, model, optimizer_model, epoch, epoch_losses, epoch_uncertainty, run_model, monitor_model,
          vis_model):
    current_time = datetime.datetime.now()
    print('\nEpoch: %d, Time: %s' % (epoch, current_time.strftime('%Y-%m-%d %H:%M:%S')))

    train_loss = 0
    prec_train_all = 0.

    for batch_idx, (inputs, targets, index) in enumerate(train_loader):
        jt.sync_all()
        model.train()

        outputs = model(inputs)
        loss = ce_loss(outputs, targets)
        optimizer_model.step(loss)

        prec_train = accuracy(outputs.data, targets.data)

        train_loss += (loss.item() * outputs.shape[0])
        prec_train_all += (prec_train.item() * outputs.shape[0])

        with jt.no_grad(no_fuse=1):
            unce = (nn.softmax(outputs, dim=1)).max()

            epoch_losses[index] = ce_loss(outputs, targets, reduce=False)
            epoch_uncertainty[index] = unce

        if (batch_idx + 1) % 50 == 0:
            print('Epoch: [%d/%d]\t'
                  'Iters: [%d/%d]\t'
                  'Loss: %.4f\t'
                  'Prec@1 %.2f' % (
                      (epoch + 1), args.epochs, batch_idx + 1, len(train_loader), (train_loss / (batch_idx + 1)),
                      prec_train))
        #打印时间戳
    starttime = datetime.datetime.now()
    print("epoch %d start time: %s", epoch, starttime.strftime('%Y-%m-%d %H:%M:%S'))                  
    monitor_model.track(epoch)
    logs = vis_model.show(epoch)

    # writer1.add_scalar('train_loss', train_loss, epoch)
    run_model.track(logs, context={'subset': 'train'})
    #打印时间戳
    endtime = datetime.datetime.now()
    print("epoch %d end time: %s", epoch, endtime.strftime('%Y-%m-%d %H:%M:%S'))
    print('epcho beihang tool 记录消耗时间：', (endtime - starttime).seconds)
    return (train_loss / len(train_loader.dataset)), (
                prec_train_all / len(train_loader.dataset)), epoch_losses, epoch_uncertainty


def extract_features(model, dataloader):
    features = []
    labels = []
    feature_extractor = copy.deepcopy(model)
    feature_extractor.linear = nn.Identity()
    with jt.no_grad(no_fuse=1):
        for inputs, targets, _ in dataloader:
            output = feature_extractor(inputs)
            features.append(output.reshape(output.shape[0], -1).numpy())
            labels.append(targets.numpy())
    return np.concatenate(features), np.concatenate(labels)


def get_output(model, dataloader):
    pre = jt.zeros(len(dataloader.dataset), args.num_classes)
    labels = jt.zeros(len(dataloader.dataset)).long()
    feature_extractor = copy.deepcopy(model)
    feature_extractor.eval()
    with jt.no_grad():
        for inputs, targets, index in dataloader:
            output = feature_extractor(inputs)

            pre[index] = output
            labels[index] = targets
    return pre, labels


# load dataset
train_loader = build_dataset(args)
args.num_classes = len(train_loader.classes)

# load model
model = build_model()
optimizer_model = jt.optim.SGD(model.parameters(), args.lr, momentum=args.momentum, weight_decay=args.weight_decay)

current_time = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
tensorboard_log_path = f'./tensorboard_log/{current_time}'

# writer1 = SummaryWriter(log_dir=(tensorboard_log_path+'/runs/epoch_model_informations'))
# writer2 = SummaryWriter(log_dir=(tensorboard_log_path+'/runs/epoch_weighting_informations'))
aim_server = os.getenv('TRACKSERVER')
print('aim_server:', aim_server)
#run_model = aim.Run(repo='aim://203.83.235.100:30058', system_tracking_interval=None, log_system_params=False)
run_model = aim.Run(repo=aim_server, system_tracking_interval=None, log_system_params=False)
run_model['id'] = os.getenv('RUNID')
run_model['hparams'] = {
    'learning_rate': args.lr,
    'batch_size': args.batch_size,
    'epochs': args.epochs,
    'momentum': args.momentum
}

def main():
    best_acc = 0
    # tensorboard_log_path = './'

    train_loss_all = []
    prec_train_all = []

    epoch_losses = jt.zeros(len(train_loader.dataset))
    epoch_uncertainty = jt.zeros(len(train_loader.dataset))

    out_file_list = []
    config_mmonitor_resnet = prepare_config()

    monitor_resnet = Monitor(model, config_mmonitor_resnet)
    vis_model = Visualization(monitor_resnet, project=config_mmonitor_resnet.keys(),
                              name=config_mmonitor_resnet.values())
    for epoch in range(args.epochs):
        adjust_learning_rate(optimizer_model, epoch)
        train_loss, prec_train, epoch_losses, epoch_uncertainty = train(train_loader, model, optimizer_model, epoch,
                                                                        epoch_losses, epoch_uncertainty, run_model,
                                                                        monitor_resnet, vis_model)

        # pre, labels = get_output(model, train_loader)

        # model.save(os.path.join('%s' % you_should_save_here, 'model_resnet18_%d.pkl' % (epoch)))

        # np.save(os.path.join('%s' % you_should_save_here, 'pre_%s.npy' % (epoch)), pre.data, allow_pickle=True,
        #         fix_imports=True)
        # np.save(os.path.join('%s' % you_should_save_here, 'labels_%s.npy' % (epoch)), labels.data, allow_pickle=True,
        #         fix_imports=True)
        print("track metric start time:", datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))
        run_model.track(train_loss, name='Training Loss', epoch=epoch)

        run_model.track(prec_train, name='Training Accuracy', epoch=epoch)
        print("track metric end time:", datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))
        # with jt.no_grad(no_fuse=1):
        #     model.eval()
        #     run_model.track(train_loss, name='Training Loss', epoch=epoch)
        #
        #     run_model.track(prec_train, name='Training Accuracy', epoch=epoch)
        #
        #     train_loss_all.append(train_loss)
        #     prec_train_all.append(prec_train)
        #
        #     plt.figure()
        #     plt.plot(train_loss_all, label='Model Loss')
        #     plt.xlabel('Epoch')
        #     plt.ylabel('Loss')
        #     plt.title('Loss vs Epoch')
        #     plt.legend()
        #     plt.grid(True)
        #
        #     plt.xticks(np.arange(len(train_loss_all)))
        #
        #     plot_filename = (
        #                 tensorboard_log_path + f'/runs/epoch_model_informations/loss_all_for_epoch_{epoch + 1}.png')
        #     # 创建文件夹
        #     os.makedirs(os.path.dirname(plot_filename), exist_ok=True)
        #     plt.savefig(plot_filename)
        #     plt.close()
        #
        #     with Image.open(plot_filename) as img:
        #         img_array = np.array(img)  # 转换为 numpy 数组
        #         aim_image = aim.Image(img_array)  # 使用 Aim 提供的 Image 类型包装图像
        #
        #     # 使用 Aim 记录图像
        #     run_model.track(
        #         aim_image,
        #         name=f'Loss all for Epoch',
        #         epoch=epoch,
        #         context={'model': 'Loss all for Epoch'}
        #     )
        #
        #     out_file_list.append(plot_filename)
        #
        #     # 绘制模型的准确率变化图
        #     plt.figure()
        #     plt.plot(prec_train_all, label='Model Accuracy')
        #     plt.xlabel('Epoch')
        #     plt.ylabel('Accuracy')
        #     plt.title('Accuracy vs Epoch')
        #     plt.legend()
        #     plt.grid(True)
        #
        #     plt.xticks(np.arange(len(prec_train_all)))
        #
        #     plot_filename = (
        #             tensorboard_log_path + f'/runs/epoch_model_informations/accuracy_all_for_epoch_{epoch + 1}.png')
        #     os.makedirs(os.path.dirname(plot_filename), exist_ok=True)
        #     plt.savefig(plot_filename)
        #     plt.close()
        #
        #     with Image.open(plot_filename) as img:
        #         img_array = np.array(img)  # 转换为 numpy 数组
        #         aim_image = aim.Image(img_array)  # 使用 Aim 提供的 Image 类型包装图像
        #
        #     # 使用 Aim 记录图像
        #     run_model.track(
        #         aim_image,
        #         name=f'Accuracy all for Epoch',
        #         epoch=epoch,
        #         context={'model': 'Accuracy all for Epoch'}
        #     )
        #
        #     out_file_list.append(plot_filename)
        #
        #     features, labels = extract_features(model, train_loader)
        #     tsne = TSNE(n_components=2, random_state=42)
        #     features_2d = tsne.fit_transform(features)
        #
        #     class_names = list(range(10))
        #     # 绘制 t-SNE 特征分布图，0 标识颜色深，1 标识颜色浅
        #     colors = plt.cm.tab10(np.linspace(0, 1, 10))
        #
        #     plt.figure(figsize=(10, 10))
        #     for class_idx in range(10):
        #         indices_0 = (labels == class_idx)
        #         # plt.scatter(features_2d[indices_0, 0], features_2d[indices_0, 1], label=f'{class_names[class_idx]} (dark)', cmap='tab10', alpha=0.9)
        #         # plt.scatter(features_2d[indices_1, 0], features_2d[indices_1, 1], label=f'{class_names[class_idx]} (light)', cmap='tab10', alpha=0.3)
        #         plt.scatter(features_2d[indices_0, 0], features_2d[indices_0, 1], color=colors[class_idx],
        #                     label=f'{class_names[class_idx]} (dark)', alpha=0.9)
        #
        #     plt.title('t-SNE Feature Distribution of ResNet32 on CIFAR-10')
        #     plt.xlabel('Dimension 1')
        #     plt.ylabel('Dimension 2')
        #     plt.grid(True)
        #     plt.legend(loc='best')
        #
        #     plot_filename = (
        #             tensorboard_log_path + f'/runs/epoch_model_informations/feature_distribution_for_epoch_{epoch + 1}.png')
        #     os.makedirs(os.path.dirname(plot_filename), exist_ok=True)
        #     plt.savefig(plot_filename)
        #     plt.close()
        #
        #     with Image.open(plot_filename) as img:
        #         img_array = np.array(img)  # 转换为 numpy 数组
        #         aim_image = aim.Image(img_array)  # 使用 Aim 提供的 Image 类型包装图像
        #
        #     # 使用 Aim 记录图像
        #     run_model.track(
        #         aim_image,
        #         name=f'feature Distribution at Epoch',
        #         epoch=epoch,
        #         context={'model': 'feature Distribution at Epoch'}
        #     )
        #
        #     out_file_list.append(plot_filename)
        #
        #     # ##############################################################################################
        #     class_names = list(range(args.num_classes))
        #     all_preds = []
        #     all_labels = []
        #
        #     for inputs, labels, _ in train_loader:
        #         # print('inputs:', labels)
        #         outputs = model(inputs)
        #         preds, _ = outputs.argmax(1)
        #         all_preds.extend(preds.data)
        #         all_labels.extend(labels.data)
        #
        #     conf_matrix = confusion_matrix(all_labels, all_preds)
        #
        #     # 绘制混淆矩阵的热力图
        #     plt.figure(figsize=(10, 8))
        #     sns.heatmap(conf_matrix, annot=True, cmap='Blues', fmt='d',
        #                 xticklabels=class_names, yticklabels=class_names)
        #     plt.xlabel('Predicted Label')
        #     plt.ylabel('True Label')
        #     plt.title('Confusion Matrix for ResNet32 on CIFAR-10')
        #
        #     plot_filename = (
        #             tensorboard_log_path + f'/runs/epoch_model_informations/prediction_distribution_for_epoch_{epoch + 1}.png')
        #     os.makedirs(os.path.dirname(plot_filename), exist_ok=True)
        #     plt.savefig(plot_filename)
        #     plt.close()
        #
        #     with Image.open(plot_filename) as img:
        #         img_array = np.array(img)  # 转换为 numpy 数组
        #         aim_image = aim.Image(img_array)  # 使用 Aim 提供的 Image 类型包装图像
        #
        #     # 使用 Aim 记录图像
        #     run_model.track(
        #         aim_image,
        #         name=f'Prediction Distribution at Epoch',
        #         epoch=epoch,
        #         context={'model': 'Prediction Distribution at Epoch'}
        #     )
        #
        #     plt.figure()
        #
        #     # print('epoch_losses:', type(epoch_losses), epoch_losses.shape, epoch_losses  )
        #     x1, y1 = np.histogram(epoch_losses.numpy(), bins=20, range=(0., epoch_losses.numpy().max().item()))
        #     plt.bar(np.arange(0., 1., 1 / 20), x1, width=0.01, facecolor='Blue')
        #
        #     plt.xlabel('Loss')
        #     plt.ylabel('Sample number')
        #     plt.title(f'Loss Distribution for Epoch')
        #     plt.legend()
        #     plt.grid()
        #
        #     plot_filename = (
        #             tensorboard_log_path + f'/runs/epoch_weighting_informations/loss_distribution_for_epoch_{epoch + 1}.png')
        #     os.makedirs(os.path.dirname(plot_filename), exist_ok=True)
        #     plt.savefig(plot_filename)
        #     plt.close()
        #
        #     with Image.open(plot_filename) as img:
        #         img_array = np.array(img)  # 转换为 numpy 数组
        #         aim_image = aim.Image(img_array)  # 使用 Aim 提供的 Image 类型包装图像
        #
        #     # 使用 Aim 记录图像
        #     run_model.track(
        #         aim_image,
        #         name=f'Loss Distribution at Epoch',
        #         epoch=epoch,
        #         context={'vnet': 'Loss Distribution at Epoch'}
        #     )
        #
        #     out_file_list.append(plot_filename)
        #
        #     plt.figure()
        #
        #     x1, y1 = np.histogram(epoch_uncertainty.numpy(), bins=20, range=(0., epoch_uncertainty.max().item()))
        #     plt.bar(np.arange(0., 1., 1 / 20), x1, width=0.01, facecolor='Blue')
        #
        #     plt.xlabel('Uncertainty')
        #     plt.ylabel('Sample number')
        #     plt.title(f'Uncertainty Distribution for Epoch {epoch + 1}')
        #     plt.legend()
        #     plt.grid()
        #
        #     plot_filename = (
        #             tensorboard_log_path + f'/runs/epoch_weighting_informations/uncertainty_distribution_for_epoch_{epoch + 1}.png')
        #     os.makedirs(os.path.dirname(plot_filename), exist_ok=True)
        #     plt.savefig(plot_filename)
        #     plt.close()
        #
        #     out_file_list.append(plot_filename)
    pre, labels = get_output(model, train_loader)

    # model.save(os.path.join('%s' % you_should_save_here, 'model_resnet18_%d.pkl' % (epoch)))

    # np.save(os.path.join('%s' % you_should_save_here, 'pre_%s.npy' % (epoch)), pre.data, allow_pickle=True,
    #             fix_imports=True)
    # np.save(os.path.join('%s' % you_should_save_here, 'labels_%s.npy' % (epoch)), labels.data, allow_pickle=True,
    #             fix_imports=True)
    # os.rename(os.path.join(you_should_save_here, f'pre_{epoch}.npy'), os.path.join(you_should_save_here, 'pre.pkl'))
    # os.rename(os.path.join(you_should_save_here, f'labels_{epoch}.npy'),
    #           os.path.join(you_should_save_here, 'labels.pkl'))

    # upload_output()


if __name__ == '__main__':
    main()


from .base_class import SingleStepQuantity

import jittor as jt

class WeightStd(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.weight
        
        return jt.std(data)



from ..extension import Extension
from ..utils.calculation import cal_cov_matrix,cal_eig

class BackwardOutputEigOfCovExtension(Extension):

    def __init__(self):
        super(BackwardOutputEigOfCovExtension, self).__init__()
        self._name = 'output_grad_eig'

    def _default(self, module, grad_input, grad_output):
        data = grad_output[0]
        cov = cal_cov_matrix(data)
        result = cal_eig(cov)
        return result.real

    def _Linear(self, module, grad_input, grad_output):
        data = grad_output[0]
        cov = cal_cov_matrix(data)
        result = cal_eig(cov)
        return result.real

    def _Conv(self, module, grad_input, grad_output):
        data = grad_output[0]
        b, c, w, h = data.shape
        assert (c > 1), "channel must > 1"
        data = data.transpose(0, 1).contiguous().view(-1, c)
        cov = cal_cov_matrix(data)
        result = cal_eig(cov)
        return result.real




import os
import json
import numpy as np
import pandas as pd
from ..figures import *
from ..io_utils import *

class LocalVisualization:
    def __init__(self,dir='./output', project='task', save_dir='./output/picture', 
                figsize=(16, 10), dpi=80, ncols=3):
        self.dir = dir
        self.task = project
        self.save_dir = save_dir
        self.data_loader = LoadTaskData(root=self.dir, task=self.task)
        self.ncols = 3
        self.figsize = figsize
        self.dpi = dpi
        self.figure = None 

    def show(self, monitor,quantity_name=None,project_name=None, data_type='monitor'):
        figures,selected_quantity_name = self._plot(monitor,quantity_name,project_name,data_type)
        if figures is not None:
            for i, fig in enumerate(figures):
                file_name = f"{selected_quantity_name[i]}_result.png"
                fig.save(file_name=file_name)
            
        
    def _select_figure(self, data):
        return LineFigure

    
    def save(self, quantity_name, project_name=None, data_type='monitor',
             save_dir='./output/picture', file_name=None, save_type='png', save_subfigures=True):
        figures = self._plot(quantity_name, project_name,data_type)
        if figures is not None:
            dir = save_dir
            if dir is None:
                dir = self.save_dir
            if not os.path.exists(dir):
                os.makedirs(dir)
            file_name = file_name
            if file_name is None:
                file_name = quantity_name + '.' + save_type
            file_path = os.path.join(dir, file_name)
            self.figure.savefig(file_path)
            if save_subfigures:
                if len(figures) > 1:
                    for figure in figures:
                        figure.save()
        self._clear_figure()

        
    def get_project_name(self):
        return self.data_loader.get_project_name()
    
    def get_quantity_name(self, project_name, data_type='monitor'):
        return self.data_loader.get_quantity_name(project_name, data_type)
    
    def _cal_gridspe(self, num):
        nrows = num // self.ncols
        add = num % self.ncols != 0
        return nrows + add
    
    def _plot(self, monitor,quantity_name,project_name=None, data_type='monitor'):
        figures = []
        selected_quantity_name = []
        pro = list(monitor.get_output().keys())[0]
        if quantity_name is None:
            for name in self.get_quantity_name(project_name, data_type):
                selected_quantity_name.append(name)
        else:
            quantity_name = pro+'_'+quantity_name
            selected_quantity_name.append(quantity_name)
        nrows = self._cal_gridspe(len(selected_quantity_name))
        print(f"selected_quantity_name:{selected_quantity_name}")
        if self.figure is None:
            fcal, frow = self.figsize
            self.figure = plt.figure(figsize=(fcal, frow*nrows), dpi=self.dpi)
        if len(selected_quantity_name) > 1:
            for i, name in enumerate(selected_quantity_name):
                data = self.data_loader.load_data(name, data_type)
                figure = self._select_figure(data)(data, name)
                figure.plot()
                figures.append(figure)
        elif len(selected_quantity_name) == 1:
            name = selected_quantity_name[0]
            data = self.data_loader.load_data(selected_quantity_name[0], data_type)
            # ax = self.figure.add_subplot()
            figure = self._select_figure(data)(data, name)
            figure.plot()
            figures.append(figure)
        else:
            return None
        return figures,selected_quantity_name
    
    def _clear_figure(self):
        self.figure = None


from abc import ABC
import jittor as jt
class Extension(ABC):
    def _get_module_extension_hook(self,module):
        try:
            hook = getattr(self,  '_' + module.__class__.__name__)
        except AttributeError:
            hook = self._default
        return hook
    @jt.no_grad()
    def __call__(self, module, input, output):
        module_hook = self._get_module_extension_hook(module) 
        if module_hook is not None:
            result = module_hook(module, input, output)
            setattr(module, self._name, result) 

    def _default(self, module, input, output):
        raise NotImplementedError
    


from .base_class import MultiStepQuantity

class WeightParamJump(MultiStepQuantity):

    def _compute_ones(self, global_step):
        data = self._module.weight
        return data
    
    def _compute(self, global_step):
        if len(self.cache) == 1:
            return 0
        jump_num = [d * d_p < 0 for d, d_p in zip(self.cache[-1], self.cache[-2])]
        return sum(jump_num)


from .base_class import SingleStepQuantity
from ...extensions import ForwardOutputExtension

import jittor as jt
# 计算零激活
class ZeroActivationPrecentage(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output[0]
        zero_precentage = (data==0).sum() * 100
        return zero_precentage

    def forward_extensions(self):
        extensions = [ForwardOutputExtension()]
        return extensions



from .base_class import SingleStepQuantity
import jittor as jt

class ResRatio2Save(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.res_ratio2
        return data 

    def forward_extensions(self):
        extensions = [] 
        return extensions



from .singlestep import *
from .multistep import *


class QuantitySelector:

    @staticmethod
    def select(quantity_name):
        if quantity_name not in globals():
            raise NotImplementedError(
                "hook not found: {}".format(quantity_name))
        quantity = globals()[quantity_name]
        return quantity



## ForwardInputExtension
from .forward_input_norm import ForwardInputNorm 
from .forward_input_mean import ForwardInputMean
from .forward_input_std import ForwardInputStd
from .forward_input_cov_stable_rank import ForwardInputCovStableRank 
from .forward_input_cov_max_eig import ForwardInputCovMaxEig
from .forward_input_cov_condition import ForwardInputCovCondition
from .forward_input_cov_condition20 import ForwardInputCovCondition20
from .forward_input_cov_condition50 import ForwardInputCovCondition50
from .forward_input_cov_condition80 import ForwardInputCovCondition80
## ForwardOutputExtension 
from .linear_dead_neuron_num import LinearDeadNeuronNum
from .zero_activation_precentage import ZeroActivationPrecentage
from .forward_output_norm import ForwardOutputNorm
from .forward_output_mean import ForwardOutputMean
from .forward_output_std import ForwardOutputStd
## BackwardInputExtension
from .backward_input_norm import BackwardInputNorm
from .backward_input_mean import BackwardInputMean
from .backward_input_std import BackwardInputStd
## BackwardOutputExtension ***
from .backward_output_norm import BackwardOutputNorm
from .backward_output_mean import BackwardOutputMean
from .backward_output_std import BackwardOutputStd
### 
from .weight_norm import WeightNorm
from .weight_std import WeightStd
from .attention_save import AttentionSave
from .res_ratio1_save import ResRatio1Save
from .res_ratio2_save import ResRatio2Save
from .weight_mean import WeightMean

__all__ = [
    'ForwardInputNorm',
    'ForwardInputMean',
    'ForwardInputStd',
    'ForwardInputCovStableRank',
    'ForwardInputCovMaxEig',
    'ForwardInputCovCondition',
    'ForwardInputCovCondition20',
    'ForwardInputCovCondition50',
    'ForwardInputCovCondition80',
    'LinearDeadNeuronNum',
    'WeightMean',
    'ZeroActivationPrecentage',
    'ForwardOutputNorm',
    'ForwardOutputMean',
    'ForwardOutputStd',
    'BackwardInputNorm',
    'BackwardInputMean',
    'BackwardInputStd',
    'BackwardOutputNorm',
    'BackwardOutputMean',
    'BackwardOutputStd',
    'WeightNorm',
    'AttentionSave',
    'ResRatio1Save',
    'ResRatio2Save',
    'WeightStd'
]



from ..quantity import Quantity
class SingleStepQuantity(Quantity):

    def _should_compute(self, global_step):
        return self._track_schedule(global_step)

    def _compute(self, global_step):
        raise NotImplementedError


from ..extension import Extension


class ForwardOutputExtension(Extension):

    def __init__(self):
        self._name = 'output'
        super(ForwardOutputExtension, self).__init__()

    def _default(self, module, input, output):
        return output

    def _Linear(self, module, input, output):
        return output

    def _Conv(self, module, input, output):
        return output






import os
import argparse
import aim
import numpy as np
import jittor as jt
import optim as optim_temp 
from MMonitor.mmonitor.monitor import Monitor  
from MMonitor.visualize import Visualization
from data import *
from jittor import nn
from jittor import models
from sklearn.cluster import KMeans

import copy
import datetime

from sklearn.manifold import TSNE
from sklearn.metrics import confusion_matrix
from PIL import Image
from c2net.context import prepare,upload_output
from jittor.dataset import Dataset



# c2net_context = prepare()
cifar_10_python_path = c2net_context.dataset_path

parser = argparse.ArgumentParser(description='PyTorch WideResNet Training')
parser.add_argument('--dataset', default='cifar10', type=str, help='dataset')
parser.add_argument('--model', default='resnet32', type=str, help='model')
parser.add_argument('--num_classes', default=10, type=int, help='the number of dataset classes')
parser.add_argument('--num_meta', type=int, default=1000)
parser.add_argument('--epochs', default=120, type=int, help='number of total epochs to run')
parser.add_argument('--batch_size', '--batch-size', default=100, type=int, help='mini-batch size (default: 100)')
parser.add_argument('--lr', '--learning-rate', default=1e-1, type=float, help='initial learning rate')
parser.add_argument('--momentum', default=0.9, type=float, help='momentum')
parser.add_argument('--nesterov', default=True, type=bool, help='nesterov momentum')
parser.add_argument('--weight-decay', '--wd', default=5e-4, type=float, help='weight decay (default: 5e-4)')
parser.add_argument('--prefetch', type=int, default=1, help='Pre-fetching threads.')
parser.add_argument('--corruption_prob', type=float, default=0.4, help='label noise')
parser.add_argument('--corruption_type', '-ctype', type=str, default='unif', help='Type of corruption ("unif" or "flip" or "flip2").')
parser.add_argument('--seed', type=int, default=1)

args = parser.parse_args()

args.cifar_10_python_path = cifar_10_python_path
args.sel_data = list(range(50000))

jt.flags.use_cuda = 1


def build_model():
    model = models.Resnet18(num_classes=args.num_classes)
    return model


class VNet(nn.Module):
    def __init__(self, input, hidden1, output):
        super(VNet, self).__init__()
        self.linear1 = nn.Linear(input, hidden1)
        self.relu1 = nn.ReLU()
        self.linear2 = nn.Linear(hidden1, output)

    def execute(self, x):
        x = self.linear1(x)
        x = self.relu1(x)
        out = self.linear2(x)
        return nn.Sigmoid()(out)


def accuracy(output, target):
    batch_size = target.shape[0]
    pred = np.argmax(output, -1)
    res = ( (pred == target).astype(float).sum() )/batch_size

    return res


def print_lr(optimizer, epoch):
    for param_group in optimizer.param_groups:
        lr = param_group['lr']
        print('\n Epoch:{:.4f}|LR:{:.4f}\n'.format(epoch,lr))
    return lr
  

def adjust_learning_rate(optimizer, epochs):
    lr = args.lr * ((0.1 ** int(epochs >= 80)) * (0.1 ** int(epochs >= 100)))  # For WRN-28-10
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


class get_meta1(Dataset):
    def __init__(self, train_data, idx_to_meta_sel):
        self.imgs = train_data.imgs
        self.transform = train_data.transform

        self.idx_to_meta_sel = idx_to_meta_sel

    def __getitem__(self, idx):
        index = self.idx_to_meta_sel[idx]
        with open(self.imgs[index][0], 'rb') as f:
            img = Image.open(f).convert('RGB')
            if self.transform:
                img = self.transform(img)
            return img, self.imgs[index][1], index

    def __len__(self):
        return len(self.idx_to_meta_sel)


def ce_loss(output, target, reduce=True):
    if len(output.shape) == 4:
        c_dim = output.shape[1]
        output = output.transpose((0, 2, 3, 1))
        output = output.reshape((-1, c_dim))
    target = target.reshape((-1, ))
    target = target.broadcast(output, [1])
    target = target.index(1) == target
    
    output = output - output.max([1], keepdims=True)
    loss = output.exp().sum(1).log()
    loss = loss - (output*target).sum(1)
    if reduce:
        return loss.mean()
    else:
        return loss


def test(model, test_loader):
    model.eval()
    correct = 0
    test_loss = 0

    for _, (inputs, targets, _) in enumerate(test_loader):
        inputs, targets = jt.array(inputs), jt.array(targets)
        outputs = model(inputs)
        test_loss += nn.cross_entropy_loss(outputs, targets).detach().item()
        predicted = np.argmax(outputs.detach(), -1) 
        correct += ( (predicted == targets.data).astype(float).sum() )

    test_loss /= len(test_loader)
    accuracy = 100. * correct / len(test_loader)

    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.4f}%)\n'.format(
        test_loss, correct, len(test_loader),
        accuracy))

    return accuracy

def prepare_config():
    config_mmonitor_resnet18 = {
        nn.BatchNorm:['ForwardOutputMean','ForwardOutputStd','ZeroActivationPrecentage','LinearDeadNeuronNum','BackOutputMean','BackOutputStd'],
        nn.Conv2d:['BackOutputMean','BackOutputStd','WeightNorm']
    }
    config_mmonitor_vnet={
        nn.Linear: ['ZeroActivationPrecentage']
        # nn.ReLU: ['ZeroActivationPercentage']
    }
    return config_mmonitor_resnet18,config_mmonitor_vnet
def train(train_loader, model, vnet, optimizer_model, optimizer_vnet, epoch, meta_lr, epoch_losses, epoch_weight, epoch_uncertainty,monitor_model,monitor_vnet,vis_model,vis_vnet,run_model):
    print('\nEpoch: %d' % epoch)

    train_loss = 0
    meta_loss = 0

    prec_train_all = 0.
    prec_meta_all = 0.

    if epoch < 30:
        for batch_idx, (inputs, targets, index) in enumerate(train_loader):
            model.train()

            outputs = model(inputs)
            loss = ce_loss(outputs, targets)
            
            optimizer_model.step(loss)

            prec_meta = 0.
            prec_train = accuracy(outputs.data, targets.data)

            train_loss += (loss.item() * outputs.shape[0])
            meta_loss += (0. * outputs.shape[0])
            prec_train_all += (prec_train.item() * outputs.shape[0])
            prec_meta_all += (prec_meta * outputs.shape[0])


            with jt.no_grad(no_fuse=1):
                unce = (nn.softmax(outputs, dim=1)).max()

                epoch_losses[index] = ce_loss(outputs, targets, reduce=False)
                epoch_uncertainty[index] = unce
                epoch_weight[index] = jt.ones( len(unce) )


            if (batch_idx + 1) % 50 == 0:
                print('Epoch: [%d/%d]\t'
                    'Iters: [%d/%d]\t'
                    'Loss: %.4f\t'
                    'MetaLoss:%.4f\t'
                    'Prec@1 %.2f\t'
                    'Prec_meta@1 %.2f' % (
                        (epoch + 1), args.epochs, batch_idx + 1, len(train_loader), (train_loss / (batch_idx + 1)),
                        (meta_loss / (batch_idx + 1)), prec_train, prec_meta))
        monitor_model.track(epoch)
        logs = vis_model.show(epoch)
        print(logs)
        run_model.track(logs,context={'subset':'train'})
        return (train_loss / len(train_loader.dataset)), (meta_loss / len(train_loader.dataset)), \
            (prec_train_all / len(train_loader.dataset)), (prec_meta_all / len(train_loader.dataset)), \
            epoch_losses, epoch_uncertainty, epoch_weight

    else:
        model.eval()
        with jt.no_grad(no_fuse=1):
            num_class_all = len( list( set(train_loader.dataset.noise_label) ) )

            all_loss = jt.zeros(len(train_loader.dataset)).float()
            noise_label_all = jt.zeros(len(train_loader.dataset)).long()

            for inputs_u, targets, index in train_loader:
                output_u  = model(inputs_u)
                all_loss[index] = ce_loss(output_u, targets, reduce=False)
                noise_label_all[index] = targets
            noise_label_all = noise_label_all.data.tolist()
            
            idx_to_meta = []
            data_list = {}
            for j in range( args.num_classes ):
                data_list[j] = [i for i, label in enumerate(noise_label_all) if label == j]


            for _, img_id_list in data_list.items():
                _, indexs = jt.topk(all_loss[img_id_list], min(10, len(img_id_list)), largest=False)
                idx_to_meta.extend(((jt.array(img_id_list))[indexs]).tolist())

            train_meta_loader = get_meta1(train_loader, idx_to_meta).set_attrs(num_workers=args.prefetch, batch_size=args.batch_size,shuffle=True)

        train_meta_loader.endless = True
        train_meta_loader_iter = iter(train_meta_loader)
        for batch_idx, (inputs, targets, index) in enumerate(train_loader):
            model.train()

            meta_model = build_model()
            meta_model.load_state_dict(model.state_dict())
            optimizer_temp = optim_temp.SGD(meta_model.parameters(), meta_lr,
                                            momentum=args.momentum, weight_decay=args.weight_decay)

            outputs = meta_model(inputs)
            cost = ce_loss(outputs, targets, reduce=False).reshape((-1,1))
            v_lambda = vnet(cost.detach())
            l_f_meta = ((cost * v_lambda).sum())/(v_lambda.detach().sum())
            optimizer_temp.step(l_f_meta)

            inputs_val, targets_val, _ = next(train_meta_loader_iter)
            y_g_hat = meta_model(inputs_val)
            l_g_meta = nn.cross_entropy_loss(y_g_hat, targets_val)
            optimizer_vnet.set_input_into_param_group((inputs_val,y_g_hat))
            optimizer_vnet.step(l_g_meta)

            outputs = model(inputs)
            cost_w = ce_loss(outputs, targets, reduce=False).reshape((-1, 1))
            
            with jt.no_grad():
                w_new = vnet(cost_w)

            loss = ((cost_w * w_new).sum())/(w_new.detach().sum())
            optimizer_model.set_input_into_param_group((inputs,outputs))
            optimizer_model.step(loss)

            prec_meta = accuracy(y_g_hat.data, targets_val.data)
            prec_train = accuracy(outputs.data, targets.data)

            train_loss += (loss.item() * outputs.shape[0])
            meta_loss += (l_g_meta.item() * outputs.shape[0])
            prec_train_all += (prec_train.item() * outputs.shape[0])
            prec_meta_all += (prec_meta.item() * outputs.shape[0])


            with jt.no_grad(no_fuse=1):
                # train_loss += loss
                # meta_loss += l_g_meta

                unce = (nn.softmax(outputs, dim=1)).max()

                epoch_losses[index] = cost_w.squeeze(1)
                epoch_uncertainty[index] = unce
                epoch_weight[index] = w_new.squeeze(1)


            if (batch_idx + 1) % 50 == 0:
                print('Epoch: [%d/%d]\t'
                    'Iters: [%d/%d]\t'
                    'Loss: %.4f\t'
                    'MetaLoss:%.4f\t'
                    'Prec@1 %.2f\t'
                    'Prec_meta@1 %.2f' % (
                        (epoch + 1), args.epochs, batch_idx + 1, len(train_loader), (train_loss / (batch_idx + 1)),
                        (meta_loss / (batch_idx + 1)), prec_train, prec_meta))
        monitor_vnet.track(epoch)
        logs_vnet = vis_vnet.show(epoch)
        run_model.track(logs_vnet,context={'subset':'train'})
        monitor_model.track(epoch)
        logs_model = vis_model.show(epoch)
        run_model.track(logs_model,context={'subset':'train'})
        return (train_loss / len(train_loader.dataset)), (meta_loss / len(train_loader.dataset)), \
            (prec_train_all / len(train_loader.dataset)), (prec_meta_all / len(train_loader.dataset)), \
            epoch_losses, epoch_uncertainty, epoch_weight


def extract_features(model, dataloader):
    features = []
    labels = []
    feature_extractor = copy.deepcopy(model)
    feature_extractor.linear=nn.Identity()
    with jt.no_grad():
        for inputs, targets, _ in dataloader:
            output = feature_extractor(inputs)
            features.append(output.reshape(output.shape[0], -1).numpy())
            labels.append(targets.numpy())
    return np.concatenate(features), np.concatenate(labels)


def get_output(model, dataloader):
    pre = jt.zeros( len(train_loader.dataset), args.num_classes )
    labels = jt.zeros( len(train_loader.dataset) ).long()
    feature_extractor = copy.deepcopy(model)
    feature_extractor.eval()
    with jt.no_grad():
        for inputs, targets, index in dataloader:
            output = feature_extractor(inputs)
            
            pre[index] = output            
            labels[index] = targets
    return pre, labels


# load dataset 
train_loader = build_dataset(args)

# load model
model = build_model()
vnet = VNet(1, 100, 1)

optimizer_model = jt.optim.SGD(model.parameters(), args.lr, momentum=args.momentum, weight_decay=args.weight_decay)
optimizer_vnet = jt.optim.Adam(vnet.parameters(), 1e-3, weight_decay=1e-4)

current_time = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
tensorboard_log_path = f'./tensorboard_log/{current_time}'



def main():
    best_acc = 0

    train_loss_all = []
    meta_loss_all = []
    prec_train_all = []
    prec_meta_all = []

    epoch_losses = jt.zeros( len(train_loader.dataset) )
    epoch_weight = jt.zeros( len(train_loader.dataset) )
    epoch_uncertainty = jt.zeros( len(train_loader.dataset) )


    out_file_list = []
    config_mmonitor_resnet,config_mmonitor_vnet = prepare_config()
    run_model = aim.Run(repo='aim://203.83.235.100:30058', system_tracking_interval=None, log_system_params=False)
    monitor_resnet = Monitor(model,config_mmonitor_resnet)
    monitor_vnet = Monitor(vnet,config_mmonitor_vnet)
    vis_model = Visualization(monitor_resnet, project=config_mmonitor_resnet.keys(), name=config_mmonitor_resnet.values())
    vis_vnet = Visualization(monitor_vnet, project=config_mmonitor_vnet.keys(), name=config_mmonitor_vnet.values())
    for epoch in range(args.epochs):
        adjust_learning_rate(optimizer_model, epoch)
        meta_lr = print_lr(optimizer_model, epoch)
        train_loss, meta_loss, prec_train, prec_meta, epoch_losses, epoch_uncertainty, epoch_weight = train(train_loader, model, vnet, optimizer_model, optimizer_vnet,epoch, meta_lr, epoch_losses, epoch_weight, epoch_uncertainty,monitor_resnet,monitor_vnet,vis_model,vis_vnet,run_model)
        
        pre, labels = get_output(model, train_loader)
if __name__ == '__main__':
    main()



from model.model import Model  # Assuming Model is defined as per the previous Jittor conversion
import jittor as jt

# Configuration
config = {
    'epoch': 100,
    'w': 224,
    'h': 224,
    'class_num': 5,
    'len': 100,
    'lr': 1e-2
}

# Create input tensor
x = jt.rand((config['len'], 3, config['w'], config['h']), requires_grad=True)

# Initialize the model
model = Model()

# Forward pass
y = model(x)

# Print the output
print(f'The running result of the current model is {y.numpy()}')



from collections import defaultdict
import jittor
import jittor.nn as nn
from MMonitor.utils.schedules import ScheduleSelector, linear
from MMonitor.utils.register import Register
from MMonitor.quantity import QuantitySelector

class Monitor:
    def __init__(self, model, configer=None):
        self.model = model
        self.params = list(model.parameters())
        self.output = defaultdict(dict)
        self.configer = configer or {}
        self.parse_module, self.parse_quantity = self._config_parser()
        
        if self.configer:
            self._register()

    def track(self, global_step):
        for quantities in self.parse_quantity.values():
            for quantity in quantities:
                quantity.track(global_step)

    def get_output(self):
        self._update_output()
        return self.output

    def clean_mem(self):
        self.output.clear()
        for quantities in self.parse_quantity.values():
            for quantity in quantities:
                quantity.clean_mem()

    def _parse_quantities(self, quantities):
        quantities_list = []
        track_schedule_list = []
        
        for item in quantities:
            if isinstance(item, str):
                quantities_list.append(item)
                track_schedule_list.append(linear())
            else:
                track_schedule_list.append(
                    linear() if len(item) == 1 else ScheduleSelector.select(item[1])
                )
                quantities_list.append(item[0])
                
        return zip(quantities_list, track_schedule_list)

    def _config_parser(self):
        parse_module = defaultdict(dict)
        parse_quantity = defaultdict(dict)
        for module_name, quantities in self.configer.items():
            import pdb
            pdb.set_trace()
            try:
                module = self._get_submodule(module_name)
                module.name = module_name
                parse_module[module_name] = module
                parse_quantity[module_name] = [QuantitySelector.select(quantity)(module, track_schedule) for quantity,  track_schedule in
                                               self._parse_quantities(quantities)]
            except (AttributeError, TypeError):
                for name, module in self.model.named_modules():
                    if self._is_module(module_name, module) and name not in parse_module.keys():
                        module.name = name
                        parse_module[name] = module
                        parse_quantity[name] = [QuantitySelector.select(quantity)(module, track_schedule) for quantity,  track_schedule in
                                               self._parse_quantities(quantities)]
        return parse_module, parse_quantity

    def _get_submodule(self, target):
        if target == "":
            return self.model
        atoms = target.split(".")
        mod = self.model
        for item in atoms:

            if not hasattr(mod, item):
                raise AttributeError(mod.__class__.__name__ + " has no attribute `" + item + "`")
            if isinstance(mod, nn.Sequential) or isinstance(mod, nn.ModuleList):
                if isinstance(item, str) and len(item) > 10:
                    item = item.split('_')[-1]
                mod = mod[int(item)]
            else:
                mod = getattr(mod, item)

            if not isinstance(mod, jittor.nn.Module):
                raise AttributeError("`" + item + "` is not an nn.Module")

        return mod

    
    def _is_module(self, module_name, module):
        if isinstance(module_name, str):
            if module.__class__.__name__ == module_name:
                return True
            else:
                return False
        elif isinstance(module_name, type):
            if isinstance(module, module_name):
                return True
            else:
                return False
        else:
            return False
    def _register(self):
        for module_name, quantities in self.parse_quantity.items():
            module = self.parse_module[module_name]
            forward_extensions = self._process_duplicate_extensions(
                [quantity.forward_extensions() for quantity in quantities]
            )
            if forward_extensions:
                Register.register_forward(module, forward_extensions)

            backward_extensions = self._process_duplicate_extensions(
                [quantity.backward_extensions() for quantity in quantities]
            )
            if backward_extensions:
                Register.register_backward(module, backward_extensions)

    def _update_output(self):
        for module_name, quantities in self.parse_quantity.items():
            for quantity in quantities:
                self.output[module_name][quantity.__class__.__name__] = quantity.get_output()

    def _process_duplicate_extensions(self, extensions):
        ext_dict = {}
        no_duplicate_ext = []

        for es in extensions:
            for extension in es:
                if type(extension) not in ext_dict:
                    no_duplicate_ext.append(extension)
                    ext_dict[type(extension)] = True

        return no_duplicate_ext



# 用于计算梯度的均值
from .base_class import SingleStepQuantity
from MMonitor.extensions import BackwardOutputExtension
import jittor as jt

class BackwardOutputMean(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output_grad
        return jt.mean(data)  # 使用Jittor的mean计算


    def backward_extensions(self):
        extensions = [BackwardOutputExtension()]
        return extensions



# 用于计算激活值的std
from .base_class import SingleStepQuantity
from MMonitor.extensions import ForwardOutputExtension
import jittor as jt

class ForwardOutputStd(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output
        return jt.std(data)  

    def forward_extensions(self):
        extensions = [ForwardOutputExtension()]
        return extensions



from .base_class import SingleStepQuantity
from ...extensions import ForwardInputEigOfCovExtension
from MMonitor.extensions.utils.calculation import cal_cov_matrix,cal_eig
import jittor as jt
import math
import numpy as np

class ForwardInputCovCondition20(SingleStepQuantity):
    def _compute(self, global_step):
        eig_values, step = getattr(self._module, 'eig_values', (None, None))
        if eig_values is None or step is None or step != global_step:
            data = self._module.input_eig_data
            cov = cal_cov_matrix(data)
            eig_values = cal_eig(cov)
            # 确保 eig_values 是一个 Jittor Var
            if not isinstance(eig_values, jt.Var):
                eig_values = jt.Var(eig_values.real)
            # 对特征值进行排序（降序）并只保留值
            eig_values = jt.sort(eig_values, descending=True)[0]  # 添加[0]来只获取排序后的值
            setattr(self._module, 'eig_values', (eig_values, global_step))
        length = len(eig_values)
        index = math.floor(length * 0.2)
        eps = 1e-7
        condition20 = eig_values[0] / (jt.abs(eig_values[index]) + eps)
        return condition20


    def forward_extensions(self):
        extensions = [ForwardInputEigOfCovExtension()]
        return extensions



from .base_class import SingleStepQuantity

import jittor as jt

class ResRatio1Save(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.res_ratio1
        return data 

    def forward_extensions(self):
        extensions = []  
        return extensions



import os
from jittor import nn
from jittor.dataset import Dataset
import wandb

from MMonitor.visualize import Visualization
from MMonitor.mmonitor.monitor import Monitor
from .loader import load_monitor_config


class MonitorWandbCallback:
    def __init__(self, monitor_config=None):
        self.monitor_config = load_monitor_config(monitor_config)
        self._initialized = False
        self._wandb = wandb

    def setup(self, args, model):
        if self._wandb is None or self.monitor_config is None:
            return
        self._initialized = True
        if wandb.run is None:
            wandb.init(project=os.getenv("WANDB_PROJECT", "jittor"), config=args)

        # Log model config
        if hasattr(model, "config") and model.config is not None:
            model_config = model.config.to_dict()
            wandb.config.update(model_config)

        # Track model parameters
        _watch_model = os.getenv("WANDB_WATCH", "false")
        if _watch_model in ("all", "parameters", "gradients"):
            wandb.watch(model, log=_watch_model)

        self.monitor = Monitor(model, self.monitor_config)
        self.vis = Visualization(self.monitor, self._wandb, project=os.getenv("WANDB_PROJECT", "jittor"))

    def on_train_begin(self, args, model):
        if self._wandb is None or self.monitor_config is None:
            return
        if not self._initialized:
            self.setup(args, model)

    def on_train_end(self):
        if self._wandb is None or self.monitor_config is None:
            return
        # Optionally, handle end of training

    def on_log(self, logs=None):
        # Optionally log custom metrics
        pass

    def on_step_end(self, global_step):
        if self._wandb is None:
            return
        if not self._initialized:
            self.setup(args, model)
        if self.monitor is not None:
            self.monitor.track(global_step)
        if self.vis is not None:
            self.vis.show(global_step)

    def on_save(self):
        # Optionally handle saving the model
        pass



from MMonitor.extensions.extension import Extension
class ForwardInputExtension(Extension):

    def __init__(self):
        self._name = 'input'
        super(ForwardInputExtension, self).__init__()

    def _default(self, module, input, output):
        return input[0]

    def _Linear(self, module, input, output):
        return input[0]

    def _Conv(self, module, input, output):
        return input[0]


import jittor as jt
from jittor import nn

from ..quantity import QuantitySelector
from .schedules import ScheduleSelector


def _get_submodule(target, model):
    if target == "":
        return model

    atoms = target.split(".")
    mod = model

    for item in atoms:
        if not hasattr(mod, item):
            raise AttributeError(mod.__class__.__name__ + " has no attribute `" + item + "`")
        if isinstance(mod, nn.Sequential) or isinstance(mod, nn.ModuleList):
            mod = mod[int(item)]
        else:
            mod = getattr(mod, item)

        if not isinstance(mod, jt.nn.Module):
            raise AttributeError("`" + item + "` is not an nn.Module")

    return mod


def _is_module(module_name, module):
    if isinstance(module_name, str):
        return module.__class__.__name__ == module_name
    elif isinstance(module_name, type):
        return isinstance(module, module_name)
    else:
        return False


def _check_name(module_name, model):
    try:
        mod = _get_submodule(module_name, model)
        return module_name
    except (AttributeError, TypeError):
        model_name = []
        for name, module in model.named_modules():
            if _is_module(module_name, module):
                model_name.append(name)
        return model_name


def _check_quantity(quantity_name):
    if isinstance(quantity_name, str):
        try:
            QuantitySelector.select(quantity_name)
            return quantity_name + '_linear(0, 0)'
        except NotImplementedError:
            return None

    track_schedule = '_'
    if len(quantity_name) == 1:
        track_schedule += 'linear(0, 0)'
    else:
        try:
            ScheduleSelector.select(quantity_name[1])
            track_schedule += quantity_name[1]
        except NotImplementedError:
            track_schedule = track_schedule + quantity_name[1] + '(error)'

    try:
        QuantitySelector.select(quantity_name[0])
        return quantity_name[0] + track_schedule
    except NotImplementedError:
        return None


def check_monitor_config(config, model):
    if isinstance(model, jt.nn.parallel.DistributedDataParallel):
        model = model.module
    msg = ""
    for model_name, quantitys in config.items():
        model_n = _check_name(model_name, model)
        msg += "(model)"
        if isinstance(model_n, str):
            msg += model_n + ":"
        elif len(model_n) == 0:
            msg += model_name + "(error):"
        else:
            msg += '|'.join(model_n) + ':'
        
        quantity_name = []
        for quantity in quantitys:
            quantity_n = _check_quantity(quantity)
            if quantity_n is None:
                if isinstance(quantity, str):
                    quantity_name.append(quantity + '(error)')
                else:
                    quantity_name.append(quantity[0] + '(error)')
            else:
                quantity_name.append(quantity_n)
        
        msg += '、'.join(quantity_name)
    return msg



from ..quantity import Quantity
from ...utils.schedules import linear
import jittor as jt

class MultiStepQuantity(Quantity):
    def __init__(self, module, track_schedule=linear()):
        super(MultiStepQuantity, self).__init__(module, track_schedule)
        self.cache = []
        
    def _should_compute(self, global_step):
        return self._track_schedule(global_step)

    def _compute(self, global_step):
        raise NotImplementedError
    
    def _compute_ones(self, global_step):
        raise NotImplementedError
    
    @jt.no_grad()  # 使用Jittor的无梯度上下文管理器
    def track(self, global_step):
        self.cache.append(self._compute_ones(global_step))
        if self._should_compute(global_step):
            result = self._compute(global_step)
            self.cache.clear()
            if result is not None:
                self._save(global_step, result)


#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from setuptools import setup, find_packages
import sys

# if sys.version_info < (3, 6):
#     sys.exit('Sorry, Python >= 3.6 is required for deepund.')

setup(
    name='MMonitor-jittor',
    version='1.1.5',
    description='A Toolkit for understanding the training of neural network',
    classifiers=[
        # 发展时期,常见的如下
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        # 开发的目标用户
        'Intended Audience :: Science/Research',
        # 许可证信息
        'License :: OSI Approved :: MIT License',
        # 目标 Python 版本
        'Programming Language :: Python :: 3.8',
        # 属于什么类型
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    setup_requires=[
    ],
    install_requires=[
        'wandb',
        'plotly',
        'seaborn',
        'transformers',
        'einops',
        'aim',
        'jittor'
    ],
    packages=find_packages(),
)



# 相对于当前层输入的梯度
from ..extension import Extension
class BackwardInputExtension(Extension):
    def __init__(self):
        self._name = 'input_grad'

    def _default(self, module, grad_input, grad_output):
        return grad_input[0]

    def _Linear(self, module, grad_input, grad_output):
        return grad_input[0]

    def _Conv(self, module, grad_input, grad_output):
        return grad_input[0]


from .forward_input_extension import ForwardInputExtension
from .forward_output_extension import ForwardOutputExtension
from .forward_input_eig_of_cov_extension import ForwardInputEigOfCovExtension


__all__ = [
    'ForwardInputExtension',
    'ForwardOutputExtension',
    'ForwardInputEigOfCovExtension',
]



from collections import defaultdict
import numpy as np
import jittor as jt
from ..utils.schedules import linear 
class Quantity:

    def __init__(self, module, track_schedule=linear()):
        self._track_schedule = track_schedule
        self._module = module
        self._output = defaultdict(dict)

    def forward_extensions(self):
        return []

    def backward_extensions(self):
        return []

    @jt.no_grad()
    def track(self, global_step):
        if self._should_compute(global_step):
            result = self._compute(global_step)
            if result is not None:
                self._save(global_step, result)

    def get_output(self):
        return self._output

    def clean_mem(self):
        self.get_output().clear()

    def _should_compute(self, global_step):
        """Return if computations need to be performed at a specific iteration."""
        raise NotImplementedError

    def _save(self, global_step, result):
        """Store computation result."""
        self._output[global_step] = self._apply_save_format(result)

    def _apply_save_format(self, value):
        """Apply formatting rules for saved data."""
        if isinstance(value, jt.Var):
            value = value.numpy()  # 转换为numpy

        elif isinstance(value, dict):
            for key, val in value.items():
                value[key] = self._apply_save_format(val)

        elif isinstance(value, list):
            for idx, val in enumerate(value):
                value[idx] = self._apply_save_format(val)

        elif isinstance(value, tuple):
            value = tuple(self._apply_save_format(val) for val in value)

        elif isinstance(value, (float, int, np.ndarray)):
            pass

        else:
            raise NotImplementedError(f"No formatting rule for type {type(value)}")

        return value

    def _compute(self, global_step):
        """Evaluate quantity at a step in training."""
        raise NotImplementedError

    def should_show(self, global_step):
        return self._track_schedule(global_step)


# 测试示例
class DemoQuantity(Quantity):

    def _should_compute(self, global_step):
        return global_step % 5 == 0  # 每5个步长计算一次

    def _compute(self, global_step):
        return jt.array([global_step, global_step ** 2])  # 示例计算





from .base_figure import *

class LineFigure(Figure):
    def __init__(self, x_data, y_data, legend=None, title=None, xlabel=None, ylabel=None, line_style='-', color=None):
        super().__init__(x_data, y_data, legend)
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.line_style = line_style
        self.color = color

    def _plot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots()
        
        ax.plot(self.x_data, self.y_data, label=self.legend, linestyle=self.line_style, color=self.color)

        if self.title:
            ax.set_title(self.title)
        if self.xlabel:
            ax.set_xlabel(self.xlabel)
        if self.ylabel:
            ax.set_ylabel(self.ylabel)
        if self.legend:
            ax.legend()
        
        ax.grid(True)

    def show(self):
        plt.show()


from .base_class import SingleStepQuantity

import jittor as jt

class WeightNorm(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.weight
        data = jt.flatten(data)
        return jt.norm(data, p=2)  



import jittor as jt
import matplotlib.pyplot as plt
import seaborn as sns
import warnings

# Set up warning filtering
warnings.filterwarnings(action='once')

# Set the fontsize and some other elements
large = 22 
med = 16
small = 12
params = {
    'axes.titlesize': large,
    'legend.fontsize': med,
    'figure.figsize': (16, 10),
    'axes.labelsize': med,
    'axes.titlesize': med,
    'xtick.labelsize': med,
    'ytick.labelsize': med,
    'figure.titlesize': large
}

# Update matplotlib parameters
plt.rcParams.update(params)
sns.set_theme()
sns.set_style("white")

# Define color and line style cycles
color_cycle = [
    '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
    '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf',
    '#FF0000', '#00FF00', '#0000FF', '#FF00FF', '#FFFF00',
    '#00FFFF', '#FF5733', '#33FF57', '#5733FF', '#FF3366',
    '#33FFC7', '#B45F04', '#96C703', '#93F5C0', '#D303D3',
    '#4C1463', '#7C4B00', '#AE7F37', '#00E2E2', '#C70039'
]

line_style_cycle = ['-', '--', '-.', ':'] * 7 + ['-', '--']

# Set the prop cycle for axes
plt.rcParams['axes.prop_cycle'] = plt.cycler(color=color_cycle) + plt.cycler(linestyle=line_style_cycle)

# Example function to plot data
def plot_data(x, y, title='Plot Title', xlabel='X-axis', ylabel='Y-axis'):
    plt.figure()
    plt.plot(x, y, label=title)
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.legend()
    plt.grid()

# Example usage with Jittor
x = jt.arange(0, 10, 0.1).numpy()  # Create an array using Jittor
y = jt.sin(x).numpy()  # Compute sine values using Jittor

plot_data(x, y, title='Sine Wave', xlabel='Time', ylabel='Amplitude')


import aim
from jittor import nn
import jittor as jt
import copy
from jittor import transform
from MMonitor.mmonitor.monitor import Monitor  
from MMonitor.visualize import Visualization
from jittor.models.resnet import *
from jittor.dataset import Dataset
from jittor.dataset import CIFAR10
from PIL import Image
import numpy as np
class corrupt_train(Dataset):
    def __init__(self, traindata):
        super().__init__()
        self.imgs = traindata.imgs
        self.transform = traindata.transform
        self.sel_data = list(range(50000))

    def __getitem__(self, idx):
        k = self.sel_data[idx]
        with open(self.imgs[k][0], 'rb') as f:
            img = Image.open(f).convert('RGB')
            if self.transform:
                img = self.transform(img)
            return img, self.imgs[k][1], k
    def __len__(self):
        return len(self.imgs)
def build_dataset(cifar10_path,batch_size=100):
    transform_train = transform.Compose([
        transform.RandomHorizontalFlip(),
        transform.RandomCrop(32),
        transform.ToTensor(),
        transform.ImageNormalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
    ])
    transform_test = transform.Compose([
        transform.ToTensor(),
        transform.ImageNormalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
    ])
    # 加载数据集
    trainset = CIFAR10(root=cifar10_path,train=True, transform=transform_train,download=False)
    trainloader = trainset.set_attrs(batch_size=batch_size, shuffle=True, num_workers=2)
    return trainloader
def adjust_learning_rate(optimizer, epochs):
    lr = 1e-1
    lr = lr * ((0.1 ** int(epochs >= 80)) * (0.1 ** int(epochs >= 100)))  # For WRN-28-10
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr
def prepare_config():
    config_mmonitor = {
        nn.BatchNorm:['ForwardOutputMean','ForwardOutputStd','ZeroActivationPrecentage','LinearDeadNeuronNum'],
        nn.BatchNorm:['BackOutputMean','BackOutputStd'],
        nn.Conv2d:['BackOutputMean','BackOutputStd']
    }
    return config_mmonitor
def ce_loss(output, target, reduce=True):
    if len(output.shape) == 4:
        c_dim = output.shape[1]
        output = output.transpose((0, 2, 3, 1))
        output = output.reshape((-1, c_dim))
    target = target.reshape((-1, ))
    target = target.broadcast(output, [1])
    target = target.index(1) == target
    
    output = output - output.max([1], keepdims=True)
    loss = output.exp().sum(1).log()
    loss = loss - (output*target).sum(1)
    if reduce:
        return loss.mean()
    else:
        return loss


def accuracy(output, target):
    batch_size = target.shape[0]
    pred = np.argmax(output, -1)
    res = ( (pred == target).astype(float).sum() )/batch_size

    return res

def train(model,train_loader,vis,monitor,optimizer,epoch,meta_lr,epoch_losses,epoch_weight,epoch_uncertainty,run_model):
    print('\nEpoch: %d' % epoch)

    train_loss = 0
    meta_loss = 0

    prec_train_all = 0.
    prec_meta_all = 0.

    if epoch < 30:
        for batch_idx, (inputs, targets) in enumerate(train_loader):
            model.train()

            outputs = model(inputs)
            loss = ce_loss(outputs, targets)
            optimizer.set_input_into_param_group((inputs,outputs))
            optimizer.step(loss)
            monitor.track(batch_idx)
            logs = vis.show(batch_idx)
            run_model.track(logs,context={'subset':'train'})
            prec_meta = 0.
            prec_train = accuracy(outputs.data, targets.data)
            train_loss += (loss.item() * outputs.shape[0])
            meta_loss += (0. * outputs.shape[0])
            prec_train_all += (prec_train.item() * outputs.shape[0])
            prec_meta_all += (prec_meta * outputs.shape[0])
            # with jt.no_grad(no_fuse=1):
            #     unce = (nn.softmax(outputs, dim=1)).max()

            #     epoch_losses[index] = ce_loss(outputs, targets, reduce=False)
            #     epoch_uncertainty[index] = unce
            #     epoch_weight[index] = jt.ones( len(unce) )


            if (batch_idx + 1) % 50 == 0:
                print('Epoch: [%d/%d]\t'
                    'Iters: [%d/%d]\t'
                    'Loss: %.4f\t'
                    'MetaLoss:%.4f\t'
                    'Prec@1 %.2f\t'
                    'Prec_meta@1 %.2f' % (
                        (epoch + 1), 120, batch_idx + 1, len(train_loader), (train_loss / (batch_idx + 1)),
                        (meta_loss / (batch_idx + 1)), prec_train, prec_meta))

        return (train_loss / len(train_loader.dataset)), (meta_loss / len(train_loader.dataset)), \
            (prec_train_all / len(train_loader.dataset)), (prec_meta_all / len(train_loader.dataset)), \
            epoch_losses, epoch_uncertainty, epoch_weight
    else:
        model.eval()
        with jt.no_grad(no_fuse=1):
            num_class_all = len( list( set(train_loader.dataset.noise_label) ) )

            all_loss = jt.zeros(len(train_loader.dataset)).float()
            noise_label_all = jt.zeros(len(train_loader.dataset)).long()

            for inputs_u, targets, index in train_loader:
                output_u  = model(inputs_u)
                all_loss[index] = ce_loss(output_u, targets, reduce=False)
                noise_label_all[index] = targets
            noise_label_all = noise_label_all.data.tolist()
            
            idx_to_meta = []
            data_list = {}
            for j in range( 10 ):
                data_list[j] = [i for i, label in enumerate(noise_label_all) if label == j]


            for _, img_id_list in data_list.items():
                _, indexs = jt.topk(all_loss[img_id_list], min(10, len(img_id_list)), largest=False)
                idx_to_meta.extend(((jt.array(img_id_list))[indexs]).tolist())

            train_meta_loader = get_meta1(train_loader, idx_to_meta).set_attrs(num_workers=args.prefetch, batch_size=args.batch_size,shuffle=True)

        train_meta_loader.endless = True
        train_meta_loader_iter = iter(train_meta_loader)
        for batch_idx, (inputs, targets, index) in enumerate(train_loader):
            model.train()

            meta_model = build_model()
            meta_model.load_state_dict(model.state_dict())
            optimizer_temp = optim_temp.SGD(meta_model.parameters(), meta_lr,
                                            momentum=args.momentum, weight_decay=args.weight_decay)

            outputs = meta_model(inputs)
            cost = ce_loss(outputs, targets, reduce=False).reshape((-1,1))
            v_lambda = vnet(cost.detach())
            l_f_meta = ((cost * v_lambda).sum())/(v_lambda.detach().sum())
            optimizer_temp.step(l_f_meta)

            inputs_val, targets_val, _ = next(train_meta_loader_iter)
            y_g_hat = meta_model(inputs_val)
            l_g_meta = nn.cross_entropy_loss(y_g_hat, targets_val)

            optimizer_vnet.step(l_g_meta)

            outputs = model(inputs)
            cost_w = ce_loss(outputs, targets, reduce=False).reshape((-1, 1))
            
            with jt.no_grad():
                w_new = vnet(cost_w)

            loss = ((cost_w * w_new).sum())/(w_new.detach().sum())

            optimizer_model.step(loss)

            prec_meta = accuracy(y_g_hat.data, targets_val.data)
            prec_train = accuracy(outputs.data, targets.data)

            train_loss += (loss.item() * outputs.shape[0])
            meta_loss += (l_g_meta.item() * outputs.shape[0])
            prec_train_all += (prec_train.item() * outputs.shape[0])
            prec_meta_all += (prec_meta.item() * outputs.shape[0])


            with jt.no_grad(no_fuse=1):

                unce = (nn.softmax(outputs, dim=1)).max()

                epoch_losses[index] = cost_w.squeeze(1)
                epoch_uncertainty[index] = unce
                epoch_weight[index] = w_new.squeeze(1)


            if (batch_idx + 1) % 50 == 0:
                print('Epoch: [%d/%d]\t'
                    'Iters: [%d/%d]\t'
                    'Loss: %.4f\t'
                    'MetaLoss:%.4f\t'
                    'Prec@1 %.2f\t'
                    'Prec_meta@1 %.2f' % (
                        (epoch + 1), args.epochs, batch_idx + 1, len(train_loader), (train_loss / (batch_idx + 1)),
                        (meta_loss / (batch_idx + 1)), prec_train, prec_meta))

        return (train_loss / len(train_loader.dataset)), (meta_loss / len(train_loader.dataset)), \
            (prec_train_all / len(train_loader.dataset)), (prec_meta_all / len(train_loader.dataset)), \
            epoch_losses, epoch_uncertainty, epoch_weight


def get_output(model, dataloader):
    pre = jt.zeros( len(train_loader.dataset), 10)
    labels = jt.zeros( len(train_loader.dataset) ).long()
    feature_extractor = copy.deepcopy(model)
    feature_extractor.eval()
    with jt.no_grad():
        for inputs, targets, index in dataloader:
            output = feature_extractor(inputs)
            
            pre[index] = output            
            labels[index] = targets
    return pre, labels

def print_lr(optimizer, epoch):
    for param_group in optimizer.param_groups:
        lr = param_group['lr']
        print('\n Epoch:{:.4f}|LR:{:.4f}\n'.format(epoch,lr))
    return lr

if __name__=='__main__':
    
    cifar_10_python_path = '/data/wlc/dataset/cifar'
    model = Resnet18(pretrained=False,num_classes=10)
    config_mmonitor = prepare_config()
    run_model = aim.Run(repo='aim://203.83.235.100:30058', system_tracking_interval=None, log_system_params=False)
    print('模型已经引入')
    batch_size = 100
    train_loader = build_dataset(cifar_10_python_path,batch_size)
    print('数据已引入')
    epoch_losses = jt.zeros( len(train_loader.dataset) )
    epoch_weight = jt.zeros( len(train_loader.dataset) )
    epoch_uncertainty = jt.zeros( len(train_loader.dataset) )
    num_epochs = 120
    optimizer = jt.optim.SGD(model.parameters(), 1e-1, momentum=0.9, weight_decay=5e-4)
    monitor = Monitor(model,config_mmonitor)
    vis = Visualization(monitor, project=config_mmonitor.keys(), name=config_mmonitor.values())
    for epoch in range(num_epochs):
        adjust_learning_rate(optimizer, epoch)
        meta_lr = print_lr(optimizer, epoch)
        train_loss, meta_loss, prec_train, prec_meta, epoch_losses, epoch_uncertainty, epoch_weight = train(model,train_loader,vis,monitor,optimizer,epoch,meta_lr, epoch_losses, epoch_weight, epoch_uncertainty,run_model)
        
        pre, labels = get_output(model, train_loader)

    